import React, { useState } from "react";
import { AxisLeft, AxisBottom } from '@vx/axis';
import { AxisRight, AxisTop } from '@vx/axis';
import { AreaStack } from '@vx/shape';
import { scaleLinear, scaleOrdinal } from '@vx/scale';
import { Group } from '@vx/group';
import { extent } from 'd3-array';
import { mathjax } from 'mathjax-full/js/mathjax';
import { TeX } from 'mathjax-full/js/input/tex';
import { SVG } from 'mathjax-full/js/output/svg';
import { browserAdaptor } from 'mathjax-full/js/adaptors/browserAdaptor';
import { RegisterHTMLHandler } from 'mathjax-full/js/handlers/html';
import { STATE } from 'mathjax-full/js/core/MathItem';
import { preprocessData, sumBase, sumStat } from "../helpers/uhepp.js";

const getMaxBin =(uhepp) => {
  let max = 0;
  const n_bins = uhepp.bins.edges.length - 1
  for (let bin_i=0; bin_i < n_bins; bin_i++) {
    uhepp.stacks.forEach((stack, stack_index) => {
      let bottom = 0
      stack.content.forEach((stack_item, si_i) => {
        const y_value = sumBase(uhepp.yields, stack_item["yield"], bin_i)
        if (y_value > 0) { bottom += y_value; }
      })
      max = Math.max(max, bottom)
    })
  }
  return max;
}

// Child iterates over stacksitems = a list of rects
// rects given freedom to highlight single bins or whole processes
const Histogram = ({
  uhepp,
  yScale,
  xScale,
  colorScale,
  highlightedBin,
  onMouseOverBin,
  children
}) => {

  const edges = uhepp.bins.rebin || uhepp.bins.edges
  const n_bins = edges.length
  let objects = []

  for (let bin_i=1; bin_i < n_bins; bin_i++) {
    uhepp.stacks.forEach((stack, stack_index) => {
      let bottom = 0
      stack.content.forEach((stack_item, si_i) => {
        const y_value = sumBase(uhepp.yields, stack_item["yield"], bin_i)
        if (stack.type == "stepfilled") {
          objects.push(<rect
            key={`rect-${stack_index}-${si_i}-${bin_i}`}
            x={xScale(edges[bin_i-1])}
            width={xScale(edges[bin_i + 1]) - xScale(edges[bin_i])}
            y={yScale(bottom + y_value)}
            height={yScale(bottom) - yScale(bottom + y_value)}
            fill={stack_item.style ? stack_item.style.color : '#1f77b4'}
            opacity={highlightedBin == bin_i ? 0.5 : 1.0}
            onMouseOver={() => onMouseOverBin(bin_i)}
            />)
        } else if (stack.type == "points") {
          const width = xScale(edges[bin_i]) - xScale(edges[bin_i - 1])
          const stat = sumStat(uhepp.yields, stack_item["yield"], bin_i)

          const stat_up = yScale(y_value + stat) - yScale(y_value)
          const stat_down = yScale(y_value) - yScale(y_value - stat)
          const gx = (xScale(edges[bin_i - 1]) + xScale(edges[bin_i])) / 2
          const gy = yScale(bottom + y_value)

          if (stat && y_value) {
            objects.push(<g transform={`translate(${gx}, ${gy})`}
              key={`p-${stack_index}-${si_i}-${bin_i}`}
              >
              <circle
                cx="0"
                cy="0"
                fill="#000"
                opacity={highlightedBin == bin_i ? 0.8 : 1.0}
                onMouseOver={() => onMouseOverBin(bin_i)}
                r="3"/>
              <line
                y1="0"
                y2="0"
                stroke="#000"
                strokeWidth="2"
                strokeOpacity={highlightedBin == bin_i ? 0.8 : 1.0}
                onMouseOver={() => onMouseOverBin(bin_i)}
                x1={-width / 2}
                x2={+width / 2} />
              <line
                y1={-stat_down}
                y2={stat_up}
                strokeWidth="2"
                stroke="#000"
                strokeOpacity={highlightedBin == bin_i ? 0.8 : 1.0}
                onMouseOver={() => onMouseOverBin(bin_i)}
                x1="0"
                x2="0" />
              </g>)
          }
        }
        if (y_value > 0) {
          bottom += y_value
        }


      })

			if (stack.type == "stepfilled") {
				const whole_stack = stack.content.map(si => si["yield"]).flat()
				const stat = sumStat(uhepp.yields, whole_stack, bin_i)
				objects.push(<rect
					key={`rect-${stack_index}-uncert-${bin_i}`}
					x={xScale(edges[bin_i - 1])}
					width={xScale(edges[bin_i]) - xScale(edges[bin_i - 1])}
					y={yScale(bottom + stat)}
					fill="url(#errorBand)"
					height={yScale(bottom - stat) - yScale(bottom + stat)}
					/>)
			}

    })
  }

  return objects
}

const adaptor = browserAdaptor();
RegisterHTMLHandler(adaptor);
function onError(math) {
  const {root, typesetRoot} = math;
  if (root.toString().substr(0,14) === 'math([merror([') {
    const merror = root.childNodes[0].childNodes[0];
    const text = merror.attributes.get('data-mjx-error') || merror.childNodes[0].childNodes[0].getText();
    adaptor.setAttribute(typesetRoot, 'data-mjx-error', text);
  }
}
function updateCSS(nodeID, text) {
  let styleNode = document.getElementById(nodeID);
  if (styleNode === null) {
    styleNode = document.createElement('style');
    styleNode.setAttribute('id', nodeID);
    document.head.appendChild(styleNode);
  }
  styleNode.innerHTML = text;
}

const tex = new TeX({packages: ['base', 'ams']});
const svg = new SVG({fontCache: 'none'});
const markErrors = [STATE.TYPESET + 1, null, onError];
const tex_html = mathjax.document('', {
  InputJax: tex,
  OutputJax: svg,
  renderActions: {
    markErrors,
  }
});

const length_metrix = (str) => {
  var short_count = (str.match(/[,.tli1: ]/g) || []).length;
  var long_count = (str.match(/[MWXZ]/g) || []).length;
  return str.length + 0.2 * long_count - 0.5 * short_count
}

const EmbeddedMathJax = ({src, posX, posY, props}) => {
  let html = tex_html;
  let node = document.createElement("div");
  const math = src.trim();
  const metrics = svg.getMetricsFor(node, true);
  const outerHTML =  adaptor.outerHTML(html.convert(math, {
          display: true,
          ... metrics
  })); 
  html.updateDocument();
  updateCSS('MATHJAX-SVG-STYLESHEET', svg.cssStyles.cssText);
	node.innerHTML = outerHTML;
	let rendered_svg = node.getElementsByTagName("svg")[0];
	const scale = 6
	const transX = rendered_svg.width.baseVal.value * scale / rendered_svg.viewBox.baseVal.width
	const transY = rendered_svg.height.baseVal.value * scale / rendered_svg.viewBox.baseVal.height
  return <g 
            transform={`matrix(${transX} 0 0 ${transY} ${posX} ${posY})`}
					  dangerouslySetInnerHTML={{__html: rendered_svg.innerHTML}} { ...props}></g>
}

const MixedText = ({children, x=0, y=0}) => {
  const embedded = `\\textsf{${children}}`
  const switched = embedded.replaceAll(/\$([^$]*)\$/g, "}$1\\textsf{")
  const cleared = switched.replaceAll("\\textsf{}", "")

  return <EmbeddedMathJax posX={x} posY={y} src={cleared} />
}

const UheppHist = ({width, height, uhepp}) => {

  const [highlightedBin, setHighlightedBin] = useState(null);
  const [showTotal, setShowTotal] = useState(false);

  const post_uhepp = Object.assign({}, uhepp, {
      yields: preprocessData({
        yields: uhepp.yields,
        old_edges: uhepp.bins.edges,
        new_edges: uhepp.bins.rebin, 
        include_underflow: uhepp.bins.include_underflow, 
        include_overflow: uhepp.bins.include_overflow, 
      })
  })

	const margin = {
		top: 10,
		bottom: 60,
		left: 80,
		right: 80,
	}

	const xMax = width - margin.left - margin.right
	const yMax = height - margin.top - margin.bottom

	const xScale = scaleLinear({
		range: [0, xMax],
		domain: extent(uhepp.bins.edges),
	})
	const yScale = scaleLinear({
		range: [yMax, 0],
		domain: [0, getMaxBin(post_uhepp) * 1.3],
	})


  let i = 0
  let legend = []
  post_uhepp.stacks.reverse().map(stack => {
      stack.content.reverse().map(stack_item => {
        if (stack.type == "points") {
            let color = stack_item.style ? stack_item.style.color : '#000'
            legend.push(<g key={`legend-${i}`}
                onMouseOver={() => setShowTotal(true)}
                onMouseOut={() => setShowTotal(false)}
                >
              <circle
                cx={width * 0.5 + 10}
                cy={25 + i * 20}
                stroke={color}
                r="3"/>
              <line
                y1={25 + i * 20}
                y2={25 + i * 20}
                stroke={color}
                strokeWidth="2"
                onMouseOver={() => setShowTotal(true)}
                onMouseOut={() => setShowTotal(false)}
                x1={width * 0.5}
                x2={width * 0.5 + 20} />
              <line
                y1={20 + i * 20}
                y2={30 + i * 20}
                strokeWidth="2"
                stroke={color}
                onMouseOver={() => setShowTotal(true)}
                onMouseOut={() => setShowTotal(false)}
                x1={width * 0.5}
                x1={width * 0.5 + 10}
                x2={width * 0.5 + 10} />
              </g>)
        } else {
          legend.push(
            <rect x={width * 0.5} y={20 + i * 20}
                  key={`legend-${i}`}
                  onMouseOver={() => setShowTotal(true)}
                  onMouseOut={() => setShowTotal(false)}
                  height={10} width={20}
                  fill={stack_item.style ? stack_item.style.color : '#1f77b4'}
                  />)
        }
        if (highlightedBin != null) {
          const y_value = sumBase(post_uhepp.yields, stack_item["yield"], highlightedBin)
          const stat = sumStat(post_uhepp.yields, stack_item["yield"], highlightedBin)

          const label = y_value.toFixed(1) + " ± " + stat.toFixed(1)
          legend.push(<text fontSize={12} textAnchor="end" x={width * 0.5 + 100} y={30 + i * 20}
                key={`legend-text-${i}`}>{label}</text>)
        } else if (showTotal) {
          const y_value = sumBase(post_uhepp.yields, stack_item["yield"])
          const stat = sumStat(post_uhepp.yields, stack_item["yield"])

          const label = y_value.toFixed(1) + " ± " + stat.toFixed(1)
          legend.push(<text fontSize={12} textAnchor="end" x={width * 0.5 + 100} y={30 + i * 20}
                key={`legend-text-${i}`}>{label}</text>)

        } else {
          const label = stack_item.label
          legend.push(
            <MixedText x={width * 0.5 + 30} y={30 + i * 20}
                  key={`legend-text-${i}`}>{label}</MixedText>)
        }
        i++;
      })
      stack.content.reverse()
  })
  post_uhepp.stacks.reverse()

	return (
  <div style={{width: '100%', margin: '0 auto'}}>
			<svg viewBox={`0 0 ${width} ${height}`}
           onMouseOut={() => setHighlightedBin(null)} >
        <style>
        rect {'{shape-rendering: crispEdges }'}
        .uhepp-brand {'{font-style: italic; font-weight: bold}'}
        .uhepp-brand, .uhepp-label {'{font-size: 20px}'}
        </style>

				<pattern id="errorBand" patternUnits="userSpaceOnUse" width="6" height="6">
					<path d="M-1,1 l1,-1 M0,6 l6,-6 M5,7 l2,-2" style={{stroke: "#666",  strokeWidth: 1}} />
				</pattern>

				<Group top={margin.top} left={margin.left}>

         <text y={30} x={13} className="uhepp-brand">ATLAS</text>
         <text y={30} x={75} className="uhepp-label">Internal</text>

         <Histogram
            uhepp={post_uhepp}
            highlightedBin={highlightedBin}
            onMouseOverBin={binIndex => setHighlightedBin(binIndex)}
            stack_index={0}
            xScale={xScale}
            yScale={yScale}
          />

        {legend}
  
		 		<AxisBottom
						scale={xScale}
						top={yMax}
						stroke={'#1b1a1e'}
						tickTextFill={'#1b1a1e'}
						tickTransform="translate(0, -8)"
					/>
        <MixedText y={yMax * 1.08} x={xMax * 0.9}>{uhepp.variable.symbol}</MixedText>

		 		<AxisBottom
						scale={xScale}
						top={yMax}
            numTicks={50}
						stroke={'#1b1a1e'}
            tickLength={4}
						tickTransform="translate(0, -4)"
					  tickComponent={(props) => <></>}
					/>
					<AxisLeft
						scale={yScale}
						top={0}
						left={0}
						label={'Events'}
						stroke={'#1b1a1e'}
						tickTextFill={'#1b1a1e'}
						tickTransform="translate(8, 0)"
					/>
					<AxisLeft
						scale={yScale}
						top={0}
						left={0}
						stroke={'#1b1a1e'}
            numTicks={50}
						tickTransform="translate(4, 0)"
            tickLength={4}
					  tickComponent={(props) => <></>}
					/>
					<AxisRight
						scale={yScale}
						top={0}
						left={xMax}
						stroke={'#1b1a1e'}
					  tickComponent={(props) => <></>}
						tickTransform="translate(-8, 0)"
					/>
					<AxisRight
						scale={yScale}
						top={0}
						left={xMax}
						stroke={'#1b1a1e'}
            numTicks={50}
					  tickComponent={(props) => <></>}
            tickLength={4}
						tickTransform="translate(-4, 0)"
					/>
					<AxisTop
						scale={xScale}
						top={0}
						stroke={'#1b1a1e'}
					  tickComponent={(props) => <></>}
						tickTransform="translate(0, 8)"
					/>
					<AxisTop
						scale={xScale}
						top={0}
						stroke={'#1b1a1e'}
            numTicks={50}
					  tickComponent={(props) => <></>}
            tickLength={4}
						tickTransform="translate(0, 4)"
					/>
				</Group>
			</svg>
  </div>
	)
}
export default UheppHist;
