
from django.contrib.auth.models import User, Group
from django.urls import reverse
from django.test import RequestFactory, TestCase
from rest_framework.test import APIClient
from uhepp_vault.models import Collection, Plot

from uhepp_vault.models import PUBLIC_LEVEL, INTERNAL_LEVEL, PRIVATE_LEVEL

class CollectionsEndpointTestCase(TestCase):
    """Test the /collections endpoint of the api"""

    fixtures = ["users.json", "collections.json"]

    def setUp(self):
        self.client = APIClient()
        self.frank = User.objects.get(username="cern-fsauerbu")
        self.karl = User.objects.get(username="cern-karl")
        self.john = User.objects.get(username="cern-john")

        self.col_public = Collection.objects.get(
            owner=self.frank,
            visibility=PUBLIC_LEVEL,
        )
        self.col_internal = Collection.objects.get(
            owner=self.frank,
            visibility=INTERNAL_LEVEL,
        )
        self.col_private = Collection.objects.get(
            owner=self.frank,
            visibility=PRIVATE_LEVEL,
        )
    
        self.col_karl = Collection.objects.get(
            owner=self.karl,
            visibility=PUBLIC_LEVEL,
        )
        self.col_john = Collection.objects.get(
            owner=self.john,
            visibility=PUBLIC_LEVEL,
        )

    # List
    def plots_list(self, collection, user=None, urls=False):
        """Base test for plot listings"""
        plot = Plot.objects.create(
            uhepp={},
            collection=collection,
        )
        plot_url = reverse("plot-detail", args=[plot.uuid])

        # If a user is given, authorize her
        if user is not None:
            self.client.force_authenticate(user)

        response = self.client.get(reverse("plot-list"))
        if urls:
            plot_urls = [plot["url"] for plot in response.json()]
            return plot_url, plot_urls

        return plot_url, response

    def test_plots_list_anon_public(self):
        """Check that anonymous users can see public plots"""
        plot_url, urls = self.plots_list(self.col_public, urls=True)
        self.assertTrue(any((plot_url in url) for url in urls))

    def test_plots_list_auth_public(self):
        """Check that authorized users can see public plots"""
        plot_url, urls = self.plots_list(
            self.col_public,
            urls=True,
            user=self.john
        )
        self.assertTrue(any((plot_url in url) for url in urls))

    def test_plots_list_self_public(self):
        """Check that the owner can see public plots"""
        plot_url, urls = self.plots_list(self.col_public,
            urls=True,
            user=self.frank
        )
        self.assertTrue(any((plot_url in url) for url in urls))

    def test_plots_list_anon_internal(self):
        """Check that anonymous users cannot see internal plots"""
        plot_url, urls = self.plots_list(self.col_internal, urls=True)
        self.assertTrue(all(plot_url not in url for url in urls))

    def test_plots_list_auth_internal(self):
        """Check that authorized users can see internal plots"""
        plot_url, urls = self.plots_list(
            self.col_internal,
            urls=True,
            user=self.john
        )
        self.assertTrue(any((plot_url in url) for url in urls))

    def test_plots_list_self_internal(self):
        """Check that the owner can see internal plots"""
        plot_url, urls = self.plots_list(
            self.col_internal,
            urls=True,
            user=self.frank
        )
        self.assertTrue(any((plot_url in url) for url in urls))

    def test_plots_list_anon_private(self):
        """Check that anonymous users cannot see private plots"""
        plot_url, urls = self.plots_list(self.col_private, urls=True)
        self.assertTrue(all(plot_url not in url for url in urls))

    def test_plots_list_auth_private(self):
        """Check that authorized users cannot see private plots"""
        plot_url, urls = self.plots_list(
            self.col_private,
            user=self.john,
            urls=True
        )
        self.assertTrue(all(plot_url not in url for url in urls))

    def test_plots_list_self_private(self):
        """Check that the owner can see private plots"""
        plot_url, urls = self.plots_list(
            self.col_internal,
            urls=True,
            user=self.frank
        )
        self.assertTrue(any((plot_url in url) for url in urls))

    # Details
    def plots_details(self, collection, user=None):
        """Base test for plot details"""
        plot = Plot.objects.create(
            uhepp=dict(metadata={"filename": "yeharr"}),
            collection=collection,
        )

        # If a user is given, authorize her
        if user is not None:
            self.client.force_authenticate(user)

        response = self.client.get(reverse("plot-detail", args=[plot.uuid]))
        return response

    def test_plots_detail_anon_public(self):
        """Check that anonymous users can view public plots"""
        response = self.plots_details(self.col_public)
        self.assertEqual(
            response.json()["uhepp"]["metadata"]["filename"],
            "yeharr"
        )

    def test_plots_detail_auth_public(self):
        """Check that auth users can view public plots"""
        response = self.plots_details(self.col_public, user=self.karl)
        self.assertEqual(
            response.json()["uhepp"]["metadata"]["filename"],
            "yeharr"
        )

    def test_plots_detail_self_public(self):
        """Check that the owner cannot view public plots"""
        response = self.plots_details(self.col_public, user=self.frank)
        self.assertEqual(
            response.json()["uhepp"]["metadata"]["filename"],
            "yeharr"
        )

    def test_plots_detail_anon_internal(self):
        """Check that anonymous users can view internal plots"""
        response = self.plots_details(self.col_internal)
        self.assertEqual(response.status_code, 404)

    def test_plots_detail_auth_internal(self):
        """Check that auth users can view internal plots"""
        response = self.plots_details(self.col_internal, user=self.karl)
        self.assertEqual(
            response.json()["uhepp"]["metadata"]["filename"],
            "yeharr"
        )

    def test_plots_detail_self_internal(self):
        """Check that the owner can view internal plots"""
        response = self.plots_details(self.col_internal, user=self.frank)
        self.assertEqual(
            response.json()["uhepp"]["metadata"]["filename"],
            "yeharr"
        )

    def test_plots_detail_anon_private(self):
        """Check that anonymous users cannot view private plots"""
        response = self.plots_details(self.col_private)
        self.assertEqual(response.status_code, 404)

    def test_plots_detail_auth_private(self):
        """Check that auth users cannot view private plots"""
        response = self.plots_details(self.col_private, user=self.karl)
        self.assertEqual(response.status_code, 404)

    def test_plots_detail_self_private(self):
        """Check that the owner can view private plots"""
        response = self.plots_details(self.col_private, user=self.frank)
        self.assertEqual(
            response.json()["uhepp"]["metadata"]["filename"],
            "yeharr"
        )

    # Details
    def plots_delete(self, collection, user=None):
        """Base test for plot details"""
        plot = Plot.objects.create(
            uhepp=dict(metadata={"filename": "yeharr"}),
            collection=collection,
        )

        # If a user is given, authorize her
        if user is not None:
            self.client.force_authenticate(user)

        response = self.client.delete(reverse("plot-detail", args=[plot.uuid]))
        return response.status_code

    def test_plots_delete_anon_public(self):
        """Check that anonymous users cannot delete public plots"""
        status_code = self.plots_delete(self.col_public)
        self.assertEqual(status_code, 401)

    def test_plots_delete_auth_public(self):
        """Check that authorized users cannot delete public plots"""
        status_code = self.plots_delete(self.col_public, user=self.john)
        self.assertEqual(status_code, 403)

    def test_plots_delete_self_public(self):
        """Check that the owner can delete public plots"""
        status_code = self.plots_delete(self.col_public, user=self.frank)
        self.assertEqual(status_code, 204)

    def test_plots_delete_self_public_deleted(self):
        """Check that the owner can delete public plots"""
        plot = Plot.objects.create(
            uhepp=dict(metadata={"filename": "yeharr"}),
            collection=self.col_public,
        )

        self.client.force_authenticate(self.frank)
        response = self.client.delete(reverse("plot-detail", args=[plot.uuid]))
        
        self.assertRaises(Plot.DoesNotExist, Plot.objects.get, uuid=plot.uuid)

    def test_plots_delete_anon_internal(self):
        """Check that anonymous users cannot delete internal plots"""
        status_code = self.plots_delete(self.col_internal)
        self.assertEqual(status_code, 404)

    def test_plots_delete_auth_internal(self):
        """Check that authorized users cannot delete internal plots"""
        status_code = self.plots_delete(self.col_internal, user=self.john)
        self.assertEqual(status_code, 403)

    def test_plots_delete_self_internal(self):
        """Check that the owner can delete internal plots"""
        status_code = self.plots_delete(self.col_internal, user=self.frank)
        self.assertEqual(status_code, 204)

    def test_plots_delete_self_internal_deleted(self):
        """Check that the owner can delete internal plots"""
        plot = Plot.objects.create(
            uhepp=dict(metadata={"filename": "yeharr"}),
            collection=self.col_internal,
        )

        self.client.force_authenticate(self.frank)
        response = self.client.delete(reverse("plot-detail", args=[plot.uuid]))
        
        self.assertRaises(Plot.DoesNotExist, Plot.objects.get, uuid=plot.uuid)

    def test_plots_delete_anon_private(self):
        """Check that anonymous users cannot delete private plots"""
        status_code = self.plots_delete(self.col_private)
        self.assertEqual(status_code, 404)

    def test_plots_delete_auth_private(self):
        """Check that authorized users cannot delete private plots"""
        status_code = self.plots_delete(self.col_private, user=self.john)
        self.assertEqual(status_code, 404)

    def test_plots_delete_self_private(self):
        """Check that the owner can delete private plots"""
        status_code = self.plots_delete(self.col_private, user=self.frank)
        self.assertEqual(status_code, 204)

    def test_plots_delete_self_private_deleted(self):
        """Check that the owner can delete private plots"""
        plot = Plot.objects.create(
            uhepp=dict(metadata={"filename": "yeharr"}),
            collection=self.col_private,
        )

        self.client.force_authenticate(self.frank)
        response = self.client.delete(reverse("plot-detail", args=[plot.uuid]))
        
        self.assertRaises(Plot.DoesNotExist, Plot.objects.get, uuid=plot.uuid)

    # Update
    def plots_update(self, collection, user=None):
        """Base test for plot details"""
        plot = Plot.objects.create(
            uhepp=dict(metadata={"filename": "yeharr"}),
            collection=collection,
        )

        # If a user is given, authorize her
        if user is not None:
            self.client.force_authenticate(user)

        response = self.client.put(
            reverse("plot-detail", args=[plot.uuid]),
            json=dict(uhepp={"metadata": {"filename": "whoops"}})
        )
        status_code = response.status_code
        response = self.client.get(reverse("plot-detail", args=[plot.uuid]))

        if response.status_code == 200:
            return status_code, response.json()["uhepp"]["metadata"]["filename"]

        return status_code, response.status_code

    def test_plots_update_anon_public(self):
        """Check that anonymous users cannot update public plots"""
        status_code, filename = self.plots_update(self.col_public)
        self.assertEqual(status_code, 401)
        self.assertEqual(filename, "yeharr")

    def test_plots_update_anon_internal(self):
        """Check that anonymous users cannot update internal plots"""
        status_code, get_status = self.plots_update(self.col_internal)
        self.assertEqual(status_code, 404)
        self.assertEqual(get_status, 404)

    def test_plots_update_auth_internal(self):
        """Check that authorized users cannot update internal plots"""
        status_code, filename = self.plots_update(self.col_internal,
                                                  user=self.john)
        self.assertEqual(status_code, 403)
        self.assertEqual(filename, "yeharr")

    def test_plots_update_auth_private(self):
        """Check that authorized users cannot update private plots"""
        status_code, get_status = self.plots_update(self.col_private,
                                                    user=self.john)
        self.assertEqual(status_code, 404)
        self.assertEqual(get_status, 404)

    def test_plots_update_self_internal(self):
        """Check that authorized owner can update internal plots"""
        status_code, filename = self.plots_update(self.col_internal,
                                                  user=self.frank)
        self.assertEqual(status_code, 400)
        self.assertEqual(filename, "yeharr")

    def test_plots_update_self_private(self):
        """Check that authorized owner can update internal plots"""
        status_code, filename = self.plots_update(self.col_private,
                                                  user=self.frank)
        self.assertEqual(status_code, 400)
        self.assertEqual(filename, "yeharr")


# Following block of code tests plot update endpoints
#
#     def test_plots_update_anon_public(self):
#         """Check that anonymous users cannot update public plots"""
#         status_code, filename = self.plots_update(self.col_public)
#         self.assertEqual(status_code, 401)
#         self.assertEqual(filename, "yeharr")
# 
#     def test_plots_update_anon_internal(self):
#         """Check that anonymous users cannot update internal plots"""
#         status_code, get_status = self.plots_update(self.col_internal)
#         self.assertEqual(status_code, 404)
#         self.assertEqual(get_status, 404)
# 
#     def test_plots_update_auth_internal(self):
#         """Check that authorized users cannot update internal plots"""
#         status_code, filename = self.plots_update(self.col_internal,
#                                                   user=self.john)
#         self.assertEqual(status_code, 403)
#         self.assertEqual(filename, "yeharr")
# 
#     def test_plots_update_auth_private(self):
#         """Check that authorized users cannot update private plots"""
#         status_code, get_status = self.plots_update(self.col_private,
#                                                     user=self.john)
#         self.assertEqual(status_code, 404)
#         self.assertEqual(get_status, 404)
# 
#     def test_plots_update_self_internal(self):
#         """Check that authorized owner can update internal plots"""
#         status_code, filename = self.plots_update(self.col_internal,
#                                                   user=self.frank)
#         self.assertEqual(status_code, 204)
#         self.assertEqual(filename, "whoops")
# 
#     def test_plots_update_self_private(self):
#         """Check that authorized owner can update internal plots"""
#         status_code, filename = self.plots_update(self.col_private,
#                                                   user=self.frank)
#         self.assertEqual(status_code, 204)
#         self.assertEqual(filename, "whoops")
# 
#     # Update collection
#     def plots_update_coll(self, init_col, new_col, user=None):
#         """Base test for plot details"""
#         plot = Plot.objects.create(
#             uhepp=dict(metadata={"filename": "yeharr"}),
#             collection=init_col,
#         )
# 
#         # If a user is given, authorize her
#         if user is not None:
#             self.client.force_authenticate(user)
# 
#         response = self.client.put(
#             reverse("plot-detail", args=[plot.uuid]),
#             json=dict(collection=new_col)
#         )
#         status_code = response.status_code
#         response = self.client.get(reverse("plot-detail", args=[plot.uuid]))
# 
#         if response.status_code == 200:
#             return status_code, response.json()["collection"]
# 
#         return status_code, response.status_code
# 
#     def test_plots_update_change_self_coll(self):
#         """Check that the owner can change the collection"""
#         status_code, collection = self.plots_update_coll(
#             self.col_internal,
#             self.col_public,
#             user=self.frank
#         )
# 
#         self.assertEqual(status_code, 200)
# 
#         new_col_url = reverse("collection-detail", args=[self.col_public])
#         self.assertIn(new_col_url, collection)
# 
#     def test_plots_update_change_coll_owner_steal(self):
#         """Check that authorized owner can update internal plots"""
#         status_code, collection = self.plots_update_coll(
#             self.col_karl,
#             self.col_john,
#             user=self.john
#         )
# 
#         self.assertEqual(status_code, 403)
# 
#         new_col_url = reverse("collection-detail", args=[self.col_karl])
#         self.assertIn(new_col_url, collection)
# 
#     def test_plots_update_change_coll_owner_gift(self):
#         """Check that authorized owner can update internal plots"""
#         status_code, collection = self.plots_update_coll(
#             self.col_john,
#             self.col_karl,
#             user=self.john
#         )
# 
#         self.assertEqual(status_code, 403)
# 
#         new_col_url = reverse("collection-detail", args=[self.col_john])
#         self.assertIn(new_col_url, collection)

    # Create
    def plots_create(self, collection, user=None):
        """Check that anonymous users cannot create public plots"""
        # If a user is given, authorize her
        if user is not None:
            self.client.force_authenticate(user)

        collection_url = reverse("collection-detail", args=[collection.id])
    
        response = self.client.post(
            reverse("plot-list"),
            dict(uhepp={}, collection=collection_url),
            format="json"
        )
        status_code = response.status_code
        return status_code

    def test_plots_create_anon_public(self):
        """Check that anonymous users cannot create public plots"""
        status_code = self.plots_create(self.col_public)
        self.assertEqual(status_code, 401)  #########################

    def test_plots_create_anon_internal(self):
        """Check that anonymous users cannot create internal plots"""
        status_code = self.plots_create(self.col_internal)
        self.assertEqual(status_code, 400)

    def test_plots_create_anon_private(self):
        """Check that anonymous users cannot create internal plots"""
        status_code = self.plots_create(self.col_private)
        self.assertEqual(status_code, 400)

    def test_plots_create_auth_public(self):
        """Check that authorized users cannot create public plots"""
        status_code = self.plots_create(self.col_public, user=self.john)
        self.assertEqual(status_code, 403)  #########################

    def test_plots_create_auth_internal(self):
        """Check that authorized users cannot create internal plots"""
        status_code = self.plots_create(self.col_internal, user=self.john)
        self.assertEqual(status_code, 403)  #########################

    def test_plots_create_auth_private(self):
        """Check that authorized users cannot create internal plots"""
        status_code = self.plots_create(self.col_private, user=self.john)
        self.assertEqual(status_code, 400)  # 404 no realted object

    def test_plots_create_self_public(self):
        """Check that the owner can create public plots"""
        status_code = self.plots_create(self.col_public, user=self.frank)
        self.assertEqual(status_code, 201)

    def test_plots_create_self_internal(self):
        """Check that the owner can create internal plots"""
        status_code = self.plots_create(self.col_private, user=self.frank)
        self.assertEqual(status_code, 201)

    def test_plots_create_self_private(self):
        """Check that the owner can create private plots"""
        status_code = self.plots_create(self.col_private, user=self.frank)
        self.assertEqual(status_code, 201)

    # Headers
    def test_plots_create_headers(self):
        """Check that create response contains the correct headers"""
        collection_url = reverse("collection-detail", args=[self.col_public.pk])
    
        self.client.force_authenticate(self.frank)
        response = self.client.post(
            reverse("plot-list"),
            dict(uhepp={}, collection=collection_url),
            format="json"
        )
        
        uuid = response.json()["uuid"]

        self.assertIn(uuid, response["location"])
        self.assertEqual(response["x-resource-uuid"], uuid)
        self.assertIn(uuid, response["link"])
        self.assertIn("ui", response["link"])

    def test_plots_detail_headers(self):
        """Check that create response contains the correct headers"""
        plot = Plot.objects.create(
            collection=self.col_public,
            uhepp={}
        )
    
        self.client.force_authenticate(self.frank)
        response = self.client.get(
            reverse("plot-detail", args=[plot.uuid]),
        )

        uuid = str(plot.uuid)
        self.assertIn(uuid, response["x-resource-uuid"])
        self.assertIn(uuid, response["link"])
        self.assertIn("ui", response["link"])
